{{
*****************************************************
* jvmComm - JVM communication with programming port *
*****************************************************
}}


CON
  'Javelin IDE constants
  kReplyBit          = $80
  kFlag              = $7E
  kEscape            = $7D
  kInterrupt         = $19
  kDebugMagic        = $1D
  kRunMagic          = $A4
  kResetWait         = 10000 '// Time taken for SX to reset.
  kAckWait           = 5000  '// The amount of time (in milliseconds) to wait for acknowledgement.
  kResetCharWait     = 3     '// The amount of time to wait before sending first char after a reset.
  kProgramWait       = 2000  '// The amount of time (in milliseconds) to wait for acknowledgement when programming.
  kStepWait          = 20000 '// The amount of time (in milliseconds) to wait for acknowledgement when stepping.
  kBaseVersion       = $50   '// Compatible base version of the firmware.
  kMaxQueueLength    = 500   '// Maximum number of characters than can be queued.
  kComNone           = 0
  kComStatus         = 1
  kComReset          = 2
  kComStep           = 3
  kComRun            = 4
  kComError          = 5
  kComStop           = 6
  kComStack          = 7
  kComConfig         = 8
  kComHeap           = 9
  kComBreakpoints    = 10
  kComProgram        = 11
  kComDownloadOK     = 12
  kComStepDone       = 13
  kComAtn            = 14
  kComStartLoad      = 15
  kComEndLoad        = kBaseVersion
  kComTerminalQuery  = 17     '// Check for data from the terminal.
  kComTerminalData   = 18     '// Data follows.
  kComTerminalNoData = 19     '// No data follows.
  kComDownloadRamOK  = 20

  'values following kComError
  kErrNone       = 0
  kErrOpcode     = 1
  kErrBreakpoint = 2
  kErrMessage    = 3   'to send message to IDE message window
  kErrException  = 4   'to send exception to IDE status window
  kErrEnd        = 5   'to signal end of java program
  kErrJvm        = 6   'to send message to IDE status window


CON
  
  '//Spin stamp debug pin assignments
  stampSOUT = 16 'serial out (pin 1 of spin stamp)
  stampSIN  = 17 'serial in  (pin 2 of spin stamp)
  stampATN  = 18 'digital in (pin 3 of spin stamp, when activated, do reboot)

  '//Propeller system pin assignments
  propSCL = 28 'external eeprom SCL
  propSDA = 29 'external eeprom SDA
  propTX  = 30 'programming output
  propRX  = 31 'programming input

  
OBJ
  sp: "jvmSerial"  'serial port driver
  jd: "jvmData"    'jvm memory access object
  
  
DAT
  atnStack    long 0[10]  'stack for monitoring ATN pin
  jemArrayPtr word 0      'for downloading jembytes into javaProg array
  rangestart  word 0
  rangelength word 0
  packet      byte 0[32]  'buffer to hold packets
  reset       byte 0
  idePort     byte 0
  loadRam     byte 0      'set to true if downloaded jembytes only to ram

  
PUB start(port,monitor)
{{ start serial driver, must only be called from main }}
  idePort := port
  case port
    1: sp.start(propRX,propTX,0,28800)
    2: sp.start(stampSIN,stampSOUT,0,28800)
  'monitor ATN pin (optional) to simulate reset
  if monitor AND port == 2
    cognew(atnReset,@atnStack)


PUB doReset
{{ send reset sequence }}
  sendByte(kComStartLoad)   'send $0F
  waitcnt(clkfreq/20 + cnt) 'wait 0.05 second
  sendByte(jd.getEndLoad)   'send device version ($50, $70, $78)
  waitcnt(clkfreq/20 + cnt) 'wait 0.05 second


PUB doDebug(timeout):value | c,i,j,chk
{{ poll communication port for debug commands }}
  if timeout == 0
    c := sp.rxcheck
  else
    c := sp.rxtime(timeout)
  case c
    -1: return -1
    
    kComStatus: 'IDE sent QueryStatus command (bytes kComStatus, )
                'IDE expects packet of 6 bytes
                '(kComStatus|kReplyBit),jvm_pc_lsb,jvm_pc_msb,activity,bytecode,checksum
                sendFlag 'resetflag
                packet[0] := kComStatus | kReplyBit
                i := jd.getJvmPC+1 'IDE subtracts 1 to compensate for the increment in the JVM
                packet[1] := i.byte[0] 'pc_lsb
                packet[2] := i.byte[1] 'pc_msb
                packet[3] := jd.getActivity & $7F  'after reset activity = kPassive (=0)
                packet[4] := jd.readJavaByte(jd.getJvmPC)
                packet[5] := ComputeChecksum(@packet,5)
                sendPacket(@packet,6)
                return 0
                      
    kComConfig: 'IDE sent QueryConfig command (bytes kComConfig, )
                'IDE expects packet of 8 bytes
                '(kComConfig|kReplyBit),memsize_lsb,memsize_msb,progsize_lsb,progsize_msb,heapbase_lsb,heapbase_msb,checksum
                sendFlag 'resetflag
                packet[0] := kComConfig | kReplyBit
                packet[1] := jd.getMemSize
                packet[2] := (jd.getMemSize >> 8)
                packet[3] := jd.getProgSize
                packet[4] := (jd.getProgSize >> 8)
                packet[5] := jd.getHeapBase
                packet[6] := (jd.getHeapBase >> 8)
                packet[7] := ComputeChecksum(@packet,7)
                sendPacket(@packet,8)
                return 0

    kComStep:   'IDE sent Step command
                'respond with (kComStep|kReplyBit), do step, send (kComStepDone|kReplyBit)
                sendByte(kComStep | kReplyBit)
                jd.setNotifyStepDone 'notify jvm to signal IDE after step done
                jd.setMustDoStep     'notify jvm to perform a step
                return 0
                    
    kComRun:    'IDE sent Run command
                'run java program at full speed, respond with (kComRun|kReplyBit)
                sendByte(kComRun | kReplyBit)
                'if NOT jd.debugger
                  jd.setActivity(jd#kRun)
                'else
                '  jd.setActivity(jd#kPassive + $80)
                return 0
                    
    kComStop:   'IDE sent Stop command
                'stop java program, respond with kComStop|kReplyBit
                jd.setActivity(jd#kPassive)
                sendByte(kComStop | kReplyBit)
                return 0
                    
    kComStack: 'IDE sent QueryStack command
                'IDE expects packet of variable length
                '(kComStack|kReplyBit),jvmSP_lsb,jvmSP_msb,jvmFP_lsb,jvmFP_msb,jvmMP_lsb,jvmMP_msb,stackbytes,checksum
                sendFlag 'resetflag
                chk := 0
                sendByte(kComStack | kReplyBit)
                chk += (kComStack | kReplyBit)

                sendByte(jd.getJvmSP) 'sp_lsb
                chk += jd.getJvmSP
                sendByte(jd.getJvmSP >> 8) 'sp_msb
                chk += (jd.getJvmSP >> 8)

                sendByte(jd.getJvmFP) 'fp_lsb
                chk += jd.getJvmFP
                sendByte(jd.getJvmFP >> 8) 'fp_msb
                chk += (jd.getJvmFP >> 8)

                sendByte(jd.getJvmMP) 'mp_lsb
                chk += jd.getJvmMP
                sendByte(jd.getJvmMP >> 8) 'mp_msb
                chk += (jd.getJvmMP >> 8)

                'send stack bytes
                i := jd.getJvmSB
                repeat
                  j := jd.readJavaByte(i) & $FF 'stack data
                  sendByte(j)
                  chk += j
                  i++
                until i == jd.getJvmSP

                chk := (chk ^ $FF) + 1
                chk &= $FF
                sendByte(chk)  'checksum
                sendFlag 'endflag  
                return 0
                    
    kComHeap:   'IDE sent QueryHeap command
                'IDE sends 5 byte packet
                'kComHeap,rangestart_msb,rangestart_lsb,rangelength_msb,rangelength_lsb
                'IDE expects packet of rangelength+4 bytes
                '(kComHeap|kReplyBit),heapaddr_msb,heapaddr_lsb,<rangelength bytes>,checksum
                'get packet
                packet[0] := c
                i := 1
                repeat while i < 5
                  c := sp.rxtime(1000)
                  if c == kEscape OR c == kFlag
                    c := sp.rxtime(100)
                    c ^= $20
                  packet[i++] := c
                sendFlag 'resetflag
                rangestart.byte[0] := packet[2]
                rangestart.byte[1] := packet[1]
                rangelength.byte[0] := packet[4]
                rangelength.byte[1] := packet[3]
                chk := 0
                sendByte(kComHeap | kReplyBit)
                chk += (kComHeap | kReplyBit)

                sendByte(jd.getJvmHP >> 8) 'hp_msb
                chk += (jd.getJvmHP >> 8)
                sendByte(jd.getJvmHP) 'hp_lsb
                chk += jd.getJvmHP

                'send requested bytes
                repeat while rangelength > 0
                  i := jd.readJavaByte(rangestart)
                  sendByte(i)
                  chk += i
                  rangestart++
                  if rangestart == jd.getMemSize
                    rangestart := 0
                  rangelength--
                
                chk := (chk ^ $FF) + 1
                chk &= $FF
                sendByte(chk)  'checksum
                sendFlag 'endflag  
                return 0
                    
    kComBreakpoints: 'IDE sent DownloadBreakpoints command
                     'IDE sends variable length packet
                     'packet[0] = kComBreakpoints
                     'packet[1] = count (number of breakpoints)
                     'count words breakpoint_lsb,breakpoint_msb (last BP word can be BP for step)
                     'checksum
                     'Handle list of breakpoints and respond with kComBreakpoints|kReplyBit
                     'get packet
                     packet[0] := c
                     chk := c
                     c := sp.rxtime(1000)
                     if c == kEscape OR c == kFlag
                       c := sp.rxtime(100)
                       c ^= $20
                     chk += c
                     jd.setBreakpoints(c) 'number of breakpoints
                     i := c + c + 1 'number of bytes to receive (c words + checksum)
                     j := 0
                     repeat i
                       c := sp.rxtime(1000)
                       if c == kEscape OR c == kFlag
                         c := sp.rxtime(100)
                         c ^= $20
                       chk += c
                       jd.setBreakpointByte(j++,c) 'breakpoints are received lsb first
                     if (chk & $FF) <> 0
                       waitcnt(clkfreq/200 + cnt) 'wait 0.005 second
                       sendByte(kComBreakpoints) 'negative response
                     else
                       waitcnt(clkfreq/200 + cnt) 'wait 0.005 second
                       sendByte(kComBreakpoints | kReplyBit) 'positive response
                     return 0
                         
    kComDownloadOK:  'IDE sent DownloadOK command
                     'if ready to receive a download, respond with kComDownloadOK|kReplyBit
                     sendByte(kComDownloadOK | kReplyBit)
                     loadRam := false
                     jemArrayPtr := 0
                     return 0

    kComDownloadRamOK: 'IDE sent DownloadRamOK command
                       'if ready to receive a download, respond with kComDownloadRamOK|kReplyBit
                       sendByte(kComDownloadRamOK | kReplyBit)
                       loadRam := true
                       jemArrayPtr := 0
                       return 0

    kComAtn:         'IDE sent ATN command (used to check for echoes)
                     'do nothing
                     return 0
                         
    kComProgram: 'IDE sent kComProgram command
                 'IDE sends 19 byte packets
                 'packet[0] = kComProgram
                 'packet[1] = $01 for first packet, $00 for following packets
                 'packet[2..17] = next 16 bytes from jemfile
                 'packet[18] = checksum
                 'If checksum ok, write packet[2..17] to eeprom and respond with kComProgram|kReplyBit
                 'This state is only left when the IDE resets this device (after downloading all packets)
                 'start downloading
                 packet[0] := c
                 'get one packet
                 i := 1
                 chk := c
                 repeat while i < 19
                   c := sp.rxtime(1000)
                   if c == kEscape OR c == kFlag
                     c := sp.rxtime(100)
                     c ^= $20
                   packet[i++] := c
                   chk += c
                 if (chk & $FF) <> 0
                   waitcnt(clkfreq/200 + cnt) 'wait 0.005 second
                   sendByte(kComProgram) 'negative response
                 else
                   if packet[1] == 1
                     jemArrayPtr := 0
                   repeat i from 2 to 17
                     jd.writeJavaByte(jemArrayPtr,packet[i])
                     jemArrayPtr++
                   if NOT loadRam
                     jd.eeBlockWrite(jemArrayPtr-16)
                   waitcnt(clkfreq/200 + cnt) 'wait 0.005 second
                   sendByte(kComProgram | kReplyBit) 'positive response
                 return 0

    kComReset: 'fake reset
               sendByte(kComReset | kReplyBit)
               waitcnt(clkfreq/200 + cnt) 'wait 0.005 second
               reset := 1
               return -1
               
    other: 'unlisted value
           
  return 0
               
PUB notifyStepDone
  sendByte(kComStepDone | kReplyBit)
  jd.clearNotifyStepDone
  jd.clearMustDoStep
  

PUB breakpoint
{{ signal IDE that jvm reached a breakpoint }}
{{ after this, the jvm enters state kPassive }}
  jd.setActivity(jd#kPassive)
  packet[0] := kComError
  packet[1] := kErrBreakpoint
  sendPacket(@packet,2)
   

PUB end
{{ signal IDE that java program has ended }}
{{ after this, the jvm enters state kPassive }}
  'jd.setActivity(jd#kPassive)
  packet[0] := kComError
  packet[1] := kErrEnd
  sendPacket(@packet,2)
  

PUB errOpcode(jembyte)
{{ send unsupported/unimplemented jembyte to IDE }}
{{ after this, the jvm enters state kPassive }}
  jd.setActivity(jd#kPassive)
  packet[0] := kComError
  packet[1] := kErrOpcode
  packet[2] := jembyte
  sendPacket(@packet,3)

  
PUB errException(classOffset,pc)
{{ send exception to IDE }}
{{ after this, the jvm enters state kPassive }}
{{ pc must point to the offending jemcode instruction }}
  jd.setActivity(jd#kPassive)
  pc += 1  'add 1 because IDE subtracts one
  packet[0] := kComError
  packet[1] := kErrException
  packet[2] := classOffset.byte[1]
  packet[3] := classOffset.byte[0]
  packet[4] := pc.byte[1]
  packet[5] := pc.byte[0]
  sendPacket(@packet,6)
  end
  

PUB status(str) | len
{{ send message to IDE status window }}
{{ use this to send jvm debug messages }}
  len := strsize(str)
  if len > 0
    sp.tx(kComError)
    sp.tx(kErrJvm)
    sendPacket(str,len)


PUB message(str,len) | c,i
{{ send message to IDE message window }}
{{ this is the native CPU.message() and should not be used for debugging the jvm }}
  if len > 0
    i := 0
    sp.tx(kComError)
    sp.tx(kErrMessage)
    repeat while len > 0
      c := jd.readJavaByte(str + 4 + i)
      sendByte(c)
      i++
      len--
    sendFlag
    

PRI sendPacket(str,len)
{{ send packet }}
  if len > 0
    repeat while len > 0
      sendByte(byte[str++])
      len--
    sendFlag  
    

{{
  What is the protocol used to send messages to and from the Sin and Sout pins?
  Debug messages sent from the Javelin conform to the following pattern: 

  0x05 0x03 <messagedata> 0x7E 

  <messagedata> is the sequence of bytes being sent by the CPU.message() native method. It is byte-stuffed.
  This means that if <messagedata> contains a byte <b> that has the value 0x7D or 0x7E then that byte is
  replaced with 0x7D <b>^0x20. i.e. 0x7D followed by the orginal byte xored with 0x20. 

  Terminal characters are sent to the Javelin using the following algorithm: 

  When it is ready to receive a byte (the Terminal.getByte() native method is called) the Javelin transmits
  0x11 out Sout. 
  If the terminal has no data to send to the Javelin it replies with 0x13. 
  If the terminal has a byte <b> to send, then it replies with 0x12 <b>. 
  After the Javelin is reset it send 0x0F followed a pause followed by 0x50 (the version of the Javelin firmware).

  Running the following simple program:

  public class HelloWorld { 
    public static void main() {
      System.out.println("Hello World"); 
    } 
  }

  will generate the following on Sout:

  0F 50 05 03 48 65 6C 6C 6F 20 57 6F 72 6C 64 7E 05 03 0D 0A 7E 05 05 7E 

  Note that the sequence 0x05 0x05 0x7E indicates that the Java program has ended.

}}

PRI sendFlag
{{ send kFlag value }}
  sp.tx(kFlag)

  
PRI sendByte(txbyte)
  if txbyte == kEscape OR txbyte == kFlag
    txbyte ^= $20
    sp.tx(kEscape)
  sp.tx(txbyte)

  
PUB getByte: value | i
{{ Native method getByte }}
  sendByte(kComTerminalQuery)
  i := sp.rxtime(1000)
  if i == kComTerminalData
    i := sp.rxtime(1000)
    return i
  return -1

  
PRI ComputeChecksum(packetAddr,len): value | checksum
{{ compute checksum for packet }}
  checksum := 0
  repeat len
    checksum += byte[packetAddr++]
  return (-checksum) & $FF


PRI CheckChecksum(packetAddr,len): YesNo | checksum
{{ check checksum for packet }}
  checksum := 0
  repeat len
    checksum += byte[packetAddr++]
  return (checksum == 0)


PUB isReset: YesNo
{{ test for reset }}
  return reset


PUB setReset
{{ set reset }}
  reset := true


PUB clearReset
{{ clear reset }}
  reset := false

  
PRI atnReset
  repeat 'keep running
    if reset == false
      repeat while INA[stampATN] == 0  'wait for rising edge  
      'sp.stop
      repeat while INA[stampATN] == 1  'wait for falling edge
      'case idePort
      '  1: sp.start(propRX,propTX,0,28800)
      '  2: sp.start(stampSIN,stampSOUT,0,28800)
      'sp.start(stampSIN,stampSOUT,0,28800)
      reset := true                    'now enter reset state

      